#include <string>
#include <iostream>
#include <sstream>
#include <bits/stdc++.h> 
#include "DBData.h" 
#include "Timer.h"

class CJitContext {

public:
    std::ostringstream gencode;
    std::ostringstream initGencode;
    Tuple* rel; 
    size_t relationLen;

    CJitContext ( size_t relationLen ) {
        this->relationLen = relationLen;
    } 

    void printCode() {
        int i = system("astyle query.cpp");
        std::ifstream t("query.cpp");
        std::stringstream buffer;
        buffer << t.rdbuf();
        std::cout << "--------------- Generated C Code: ---------------" << std::endl 
                  << buffer.str()
                  << "-------------------------------------------------" << std::endl;
        if ( i != 0) {
            std::cout << "astyle not found. Install astyle for formatted code." << std::endl;
        }
    }


    std::string getFramedCode() {
        std::ostringstream oss;
        oss << "#include <iostream>" << std::endl
           << "#include <fstream>" << std::endl
           << "#include <iomanip>" << std::endl
           << "#include <chrono>" << std::endl
           << "#include <unistd.h>" << std::endl
           << "#include \"DBData.h\"" << std::endl
           << "#include \"Timer.h\"" << std::endl
           << "int main() {" << std::endl
           << "    // load or generate relation data" << std::endl
           << "    const char* dbFile = \"db.dat\";" << std::endl
           << "    Relation relation;" << std::endl
           << "    if ( !loadData ( &relation, dbFile, " << relationLen << ") ) {" << std::endl
           << "        std::cout << \"Generating data..\" << std::endl;" << std::endl
           << "        genData ( &relation, dbFile, " << relationLen << ");" << std::endl
           << "    }" << std::endl
           << "    Relation relationOut = allocateRelation (" << relationLen << ");" << std::endl
           << "    size_t numOut=0;" << std::endl
           << "    Timer tCJit = Timer();" << std::endl
           << std::endl
           << "    // query code start" << std::endl
        // your code will be inserted here
           << initGencode.str() << std::endl
        // and here 
           << gencode.str() << std::endl
           << "    // query code end" << std::endl
           << std::endl
           << "    // pass timing result to WeeDB binary" << std::endl
           << "    double time = tCJit.get();" << std::endl
           << "    std::ofstream out(\"cjitTime.log\");" << std::endl
           << "    out << time;" << std::endl  
           << "    out.close();" << std::endl
           << "    // print relation" << std::endl
           << "    relationOut.len = numOut;" << std::endl
           << "    printRelation ( relationOut );" << std::endl
           << "    return 0;" << std::endl       
           << "}" << std::endl;
        return oss.str();
    }

    void addSampleCode() {
        std::ostringstream os;
        os << "relationOut.r[0]=123;" << std::endl
           << "numOut=1;" << std::endl;
        gencode << os.str();
    }

    bool compile ( bool doPrintCode=true ) {
        std::string code = getFramedCode();
        std::ofstream out("query.cpp");
        out << code;
        out.close();
        if ( doPrintCode ) {
            printCode ();
        }
        std::string compileCommand = "g++ -O3 DBData.o query.cpp -o qbin.out";
        std::cout << compileCommand << std::endl;
        int i = system(compileCommand.c_str());
        if ( i != 0) return false;
    	return true;
    }
   
    void execute() {
        int i = system("./qbin.out");
        if ( i != 0 ) {
            std::cout << "Execute command failed." << i << std::endl;
        }
    }
    
    void addInit ( std::string newcode ) {
        initGencode << newcode;
    }
    
    void addInitLine ( std::string newcode ) {
        initGencode << newcode << std::endl;
    }

    void add ( std::string newcode ) {
        gencode << newcode;
    }
    
    void addLine ( std::string newcode ) {
        gencode << newcode << std::endl;
    }
};

